#include <avr/io.h>
#include <avr/interrupt.h>

#include "lights.h"

#define AMBIENT_ADC N_PWMLEDS
#define BATTERY_ADC (N_PWMLEDS + 1)
#define ADC1_GAIN20 (N_PWMLEDS + 2)

#define NUM_ADCS 6
volatile static unsigned char current_adc;
static uint16_t adc_sum;
static unsigned char sum_shift;
static unsigned char adc_vals;
#define ADC1_GAIN20_OFFSET_SHIFT	6
static uint16_t adc1_gain20_offset;

static void inline setup_mux(unsigned char n)
{
	/* ADC numbering: PWM LEDs first, then ambient light sensor, battery sensor */
	switch (n) {
	case 0: // pwmled 1: 1.1V, ADC0,1 (PA0,1), gain 20
		ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX1) | _BV(MUX0);
		sum_shift = PWMLED_ADC_SHIFT;
		break;
	case 1: // pwmled 2: 1.1V, ADC2,1 (PA2,1), gain 20
		ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
		sum_shift = PWMLED_ADC_SHIFT;
		break;
	case 2: // pwmled 3: 1.1V, ADC4 (PA5), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2);
		sum_shift = PWMLED_ADC_SHIFT;
		break;
	case 3: // ambient light: 1.1V, ADC5 (PA6), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2) | _BV(MUX0);
		sum_shift = 0; // 1 measurement
		break;
	case 4: // batt voltage: 1.1V, ADC6 (PA7), single-ended
		ADMUX = _BV(REFS1) | _BV(MUX2) | _BV(MUX1);
		sum_shift = 0; // 1 measurement
		break;
	case 5: // gain stage offset: 1.1V, ADC1,1, gain 20
		ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX2) | _BV(MUX0);
		sum_shift = 0; // 1 measurement
		break;
	}

	adc_sum = 0;
	adc_vals = 1 << sum_shift;
}

static void start_next_adc()
{
	if (current_adc > 0)
		current_adc--;
	else
		// TODO: kick the watchdog here.
		current_adc = NUM_ADCS-1;

	// set up mux, start one-shot conversion
	setup_mux(current_adc);
	ADCSRA |= _BV(ADSC);
}

void init_adc()
{
	unsigned char i;
	current_adc = NUM_ADCS;

	ADCSRA = _BV(ADEN)			// enable
		| _BV(ADPS1) | _BV(ADPS0)	// CLK/8 = 125 kHz
		// | _BV(ADPS2)			// CLK/16 = 62.5 kHz
		;
	// ADCSRB |= _BV(GSEL); // gain 8 or 32

	// Disable digital input on all bits used by ADC
	DIDR0 = _BV(ADC0D) | _BV(ADC1D) | _BV(ADC2D)
		| _BV(ADC4D) | _BV(ADC5D) | _BV(ADC6D);

	// 1.1V, ADC1,1, gain 20
	ADMUX = _BV(REFS1) | _BV(MUX3) | _BV(MUX2) | _BV(MUX0);
	ADCSRA |= _BV(ADSC);

	/* Do first conversion and drop the result */
	while ((ADCSRA & _BV(ADIF)) == 0)
		;
	ADCSRA |= _BV(ADIF); // clear the IRQ flag

	adc1_gain20_offset = 0;

	for (i = 0; i < (1 << ADC1_GAIN20_OFFSET_SHIFT); i++) {
		ADCSRA |= _BV(ADSC);

		while ((ADCSRA & _BV(ADIF)) == 0)
			;
		adc1_gain20_offset += ADCW
			- (adc1_gain20_offset >> ADC1_GAIN20_OFFSET_SHIFT);

		ADCSRA |= _BV(ADIF); // clear the IRQ flag
	}

	ADCSRA |= _BV(ADIE); // enable IRQ

	start_next_adc();
}

void susp_adc()
{
	ADCSRA = 0;
	DIDR0 = 0;
}

ISR(ADC_vect) { // IRQ handler
	uint16_t adcval = ADCW;

	if (adc_vals)
		// start the next conversion immediately
		ADCSRA |= _BV(ADSC);

	if (adc_vals < (1 << sum_shift))
		 // drop the first conversion, use all others
		 adc_sum += adcval;

	if (adc_vals) {
		adc_vals--;
		return;
	}

	// Now handle the (1 << sum_shift) measurements

	adcval = adc_sum >> sum_shift;

	if (current_adc == ADC1_GAIN20) {
		// running average
		adc1_gain20_offset += adcval
			- (adc1_gain20_offset >> ADC1_GAIN20_OFFSET_SHIFT);
	} else if (current_adc == 0 || current_adc == 1) {
		uint16_t offset = adc1_gain20_offset
			>> (ADC1_GAIN20_OFFSET_SHIFT - sum_shift);
		if (adc_sum > offset)
			adc_sum -= offset;
		else
			adc_sum = 0;
	}

	if (current_adc < N_PWMLEDS)
		pwmled_adc(current_adc, adc_sum);
	if (current_adc == AMBIENT_ADC)
		ambient_adc(adcval);
	if (current_adc == BATTERY_ADC)
		battery_adc(adcval);
	
	start_next_adc();
}

